@@ -1,6 +1,7 @@
 import os, requests
 from IPython.display import Markdown, display
 import sqlite3
+from sqlalchemy import column
 from tabulate import tabulate
 from Levenshtein import _levenshtein
 
@@ -40,6 +41,26 @@ class Util:
         attributes = result.field_names
         return attributes, result_list
 
+    def check_row_solution( column_total_score,
+                            partial_score_exact, 
+                            row_index, 
+                            sub_solution, 
+                            sub_student_dict,
+                            threshold = 0.8
+    ):
+        for y in range(len(sub_solution)):
+            check = Util.levenshtein_str_callback(sub_solution[y], sub_student_dict[y])
+            print(f"   #{row_index:02d}:  expecting '{sub_solution[y]}', got '{sub_student_dict[y]}'")
+            if threshold <= check:
+                column_total_score += partial_score_exact
+                print(f"         ✔ match of {check*100:.2f}%, +{partial_score_exact:.2f} pts")
+            else:
+                print(f"         ❌ match of {check*100:.2f}% below threshold of {threshold*100:.0f}%, sorry.")
+            row_index += 1
+
+        return column_total_score, row_index
+
+
     # check sql solution
     def check_sql_solution(query_text,
                            student_dict,
@@ -47,35 +68,66 @@ class Util:
                            no_of_points,
                            partial_score_exact,
                            partial_score_keywords,
-                           partial_score_points):
+                           partial_score_points,
+                           levsthn_threshold = 0.8,
+                           ):
         score = 0
         try:
+            # sanitize inputs
             solution = {Util.str_sanitize(k): v for k, v in solution.items()}
             student_dict = {Util.str_sanitize(k): v for k, v in student_dict.items()}
+
+            column_index = 0
+
+            assert no_of_points != 0, "no_of_points cannot be zero"
+            assert partial_score_exact != 0, "partial_score_exact cannot be zero"
             # check for exact match
             for x in solution:
-                print("checking columns: ", x, " | ", solution[x])
+                row_index = 0
+                column_total_score = 0
+                expected_column_score = len(solution[x]) * partial_score_exact
+                print(f">  checking column #{column_index}: '{x}'")
+                print(f">  expecting to score {expected_column_score:.2f} pts in this column")
+
                 for y in range(len(solution[x])):
-                    print("   checking cell:", y, " | ", solution[x][y])
                     check = Util.levenshtein_str_callback(solution[x][y], student_dict[x][y])
-                    if 0.8 <= check:
-                        score += partial_score_exact
-                    print("    > ", check, ", score: ", round(score, 4))
+                    print(f"   #{row_index:02d}:  expecting '{solution[x][y]}', got '{student_dict[x][y]}'")
+                    if levsthn_threshold <= check:
+                        column_total_score += partial_score_exact
+                        print(f"         ✔ match of {check*100:.2f}%, +{partial_score_exact:.2f} pts")
+                    else:
+                        print(f"         ❌ match of {check*100:.2f}% below threshold of {levsthn_threshold*100:.0f}%, sorry.")
+                    row_index += 1
+
+                column_total_score_percent = column_total_score / expected_column_score
+                print(f">  col#{column_index} ('{x}') scored {column_total_score:.2f}/{expected_column_score:.2f} = {column_total_score_percent*100:.2f}% pts.")
+                score += column_total_score
+                print(f"»  current score: {score:.4f}\n")
+                column_index += 1
         except Exception as e:
-            print('exception caught, aborting')
+            print(' ‼ exception caught, aborting')
             print(f'  {type(e).__name__} {e} ')
             if type(e).__name__ == "KeyError":
                 print("  ... this error message likely means that a column name is misspelled or that a column was not found.")
             pass  # catch, return
-        else:
+        finally:
             # if the solution isn't exact, apply other rules
             # = check for keywords in SQL.
             if round(score, 2) < no_of_points:
-                score = 0
-                print('partial score')
+                partial_score = 0
+                print(f'> Score for this task {score:.2f} is less than {no_of_points}, applying partial scoring.')
                 for i in partial_score_keywords:
                     if i in query_text.upper():
-                        score += partial_score_points
+                        print(f'  ✔ found keyword {i.rjust(10)}, +{partial_score_points:.2f} pts')
+                        partial_score += partial_score_points
+                    else:
+                        print(f'  ❌ keyword {i.rjust(10)} not found.')
+                partial_score_percent = score / no_of_points
+                print(f'> Partial scoring yielded {partial_score:.2f}/{no_of_points:.2f} = {partial_score_percent*100:.2f}% pts')
+            if partial_score > score:
+                print(f'> Awarding partial points ({partial_score:.2f}) instead of direct score ({score:.2f})')
+                score = partial_score
+        print(f'\n= Final score: {score:.2f}')
         return score
 
     # count solution cells in result array
